/* doprnt.c - pmmi strfmt primitive
 * Copyright (C) 2000 Tachibana Eriko
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

/* based on libc-1.2.20 src/stdio/_doprnt.c */

/* System headers */
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* Local headers */
#include "pmmi.h"

/* Macros */
#define DEFAULT_PRECISION 6	/* ftHg̕\x */

/* Macros */
#define UPPER_CHARSET "0123456789ABCDEF" /* 啶 */
#define LOWER_CHARSET "0123456789abcdef" /*  */

/* Macros */
#define _P_SHRTINT  0x001	/* short integer [h */
#define _P_LONGINT  0x002	/* long integer [h */
#define _P_LLNGINT  0x004	/* long long integer [h */
#define _P_LONGDBL  0x008	/* long double [h */
#define _P_SHARPEXT 0x010	/* # gtH[}bg̃[h */
#define _P_RUGLEFT  0x020	/* l߃[h */
#define _P_ZEROPAD  0x040	/* [l߃[h */
#define _P_LHEXPFIX 0x080	/* 16ivtBNX */
#define _P_UHEXPFIX 0x100	/* 啶16ivtBNX */
#define _P_OCTAPFIX 0x200	/* 8ivtBNX */
#define _P_BINAPFIX 0x400	/* 2ivtBNX */

/* Macros */
#define DEDIGIT(x) ((x) - '0')
#define ENDIGIT(x) ((x) + '0')

/* Macros */
#ifdef fputc
#undef fputc
#endif
#define fputc(c, fp)	append (c, bufp, sizep)

/* File scope functions */
static char *format_L (unsigned long long uval, int flag,
		       int precision, int base, int *size, char *charset);
static int format_last (char** bufp, int* sizep, char *ptr, int flag,
			int sign, int width, int precision, int bodysize);
static void append (int c, char** bufp, int* sizep);


/* Functions */
void doprnt (const char *format, char** bufp, int* sizep)
{
    char buf;					/* %c pobt@ */
    char *ptr, *tmp;				/* e|̃|C^ */
    int ch;					/* e|̕ */
    int count;					/* o͕ */
    int flag;					/* \[htO */
    int n;					/* e|̐l */
    int precision;				/* \x */
    int sign;					/*  */
    int size;					/* ̒ */
    int width;					/* tB[h */
    long long lval;				/* ̒l */
    unsigned long long uval;			/* ̐̒l */
    const unsigned char *fmt;			/* tH[}bg͗p */

#define GETARG() \
    ((flag & _P_SHRTINT) ? (short) popi () : popi ())

#define UGETARG() \
    ((flag & _P_SHRTINT) ? (unsigned short) popi () : popi ())

    /* tH[}bg͂ */
    for (fmt = format, count = 0;; fmt++) {
	Straw* s;

	/* ʂ̓̕oCpX */
	while ((ch = *fmt++) && (ch != '%')) {
	    fputc (ch, fp);
	    count++;
	}

	/*  */
	fmt--;

	/* null ɓBȂAI */
	if (ch == '\0')
	    {
		(*bufp)[*sizep] = '\0';
		return;
	    }

	/* parser p̃tO */
	flag = 0;
	width = 0;
	precision = -1;
	sign = '\0';

    loop:

	/* ̕o */
	switch (*++fmt) {

	case '\0':
	    (*bufp)[*sizep++] = '\0';
	    return;

	/*
	 * ̃tH[}bǵA\[hw肷....
	 */

	case ' ':
	    /* space [h + [h override łȂ */
	    if (sign == '\0')
		sign = ' ';
	    goto loop;
	case '#':
	    /* # g\[h */
	    flag |= _P_SHARPEXT;
	    goto loop;
	case '-':
	    /* l߃[h */
	    flag |= _P_RUGLEFT;
	    goto loop;
	case '+':
	    /* + [h space [h override ł */
	    sign = '+';
	    goto loop;
	case '0':
	    /* [l߃[h */
	    if (sign != '-' && precision <= 0)
		flag |= _P_ZEROPAD;
	    goto loop;
	case 'h':
	    /*  short integer ł */
	    flag |= _P_SHRTINT;
	    goto loop;
	case 'l':
	    /*  long int ł */
	    flag |= _P_LONGINT;
	    goto loop;
	case 'L':
	    /*  long long int 邢 long double ł */
	    flag |= _P_LLNGINT | _P_LONGDBL;
	    goto loop;

	/*
	 * ̃tH[}bg͐xwł....
	 */

	case '.':
	    if (*++fmt == '*')
		n = popi ();
	    else {
		n = 0;
		while (isdigit (*fmt))
		    n = 10 * n + DEDIGIT (*fmt++);
		fmt--;
	    }
	    precision = n;
	    goto loop;

	/*
	 * ̃tH[}bg̓tB[h̎wł....
	 */

	case '*':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    if (*fmt == '*') {
		if ((width = popi ()) < 0) {
		    width = -width;
		    flag |= _P_RUGLEFT;
		}
	    }
	    else {
		n = 0;
		do
		    n = 10 * n + DEDIGIT (*fmt);
		while (isdigit (*++fmt));
		width = n;
		fmt--;
	    }
	    goto loop;

	/*
	 * 16iA8i\... yу|C^`
	 */

	case 'X':
	    uval = UGETARG ();
	    if ((flag & _P_SHARPEXT) && uval != 0)
		flag |= _P_UHEXPFIX;
	    ptr = format_L (uval, flag, precision, 16, &size, UPPER_CHARSET);
	    count += format_last (bufp, sizep, ptr, flag, '\0', width, precision, size);
	    break;

	case 'p':
	    if (precision == -1)
		precision = 8;

	case 'x':
	    uval = UGETARG ();
	    if ((flag & _P_SHARPEXT) && uval != 0)
		flag |= _P_LHEXPFIX;
	    ptr = format_L (uval, flag, precision, 16, &size, LOWER_CHARSET);
	    count += format_last (bufp, sizep, ptr, flag, '\0', width, precision, size);
	    break;

	case 'o':
	    uval = UGETARG ();
	    if ((flag & _P_SHARPEXT) && uval != 0)
		flag |= _P_OCTAPFIX;
	    ptr = format_L (uval, flag, precision, 8, &size, LOWER_CHARSET);
	    count += format_last (bufp, sizep, ptr, flag, '\0', width, precision, size);
	    break;

	case 'b':
	    uval = UGETARG ();
	    if ((flag & _P_SHARPEXT) && uval != 0)
		flag |= _P_BINAPFIX;
	    ptr = format_L (uval, flag, precision, 2, &size, LOWER_CHARSET);
	    count += format_last (bufp, sizep, ptr, flag, '\0', width, precision, size);
	    break;

	/*
	 * \....
	 */

	case 'c':
	    buf = (char) popi ();
	    count += format_last (bufp, sizep, &buf, flag, '\0', width, precision, 1);
	    break;

	/*
	 * \....
	 */

	case 'Z':
	case 'U':
	    flag |= _P_LONGINT;
	case 'u':
	    uval = UGETARG ();
	    ptr = format_L (uval, flag, precision, 10, &size, LOWER_CHARSET);
	    count += format_last (bufp, sizep, ptr, flag, '\0', width, precision, size);
	    break;

	case 'D':
	case 'I':
	    flag |= _P_LONGINT;
	case 'd':
	case 'i':
	    if ((lval = GETARG ()) < 0) {
		lval = -lval;
		sign = '-';
	    }
	    ptr = format_L (lval, flag, precision, 10, &size, LOWER_CHARSET);
	    count += format_last (bufp, sizep, ptr, flag, sign, width, precision, size);
	    break;

	/*
	 * \....
	 */

	case 's':
	    s = pop ();
	    ptr = s->str;

	    /* \xATCY */
	    size = (precision < 0) ? strlen (ptr) :
		(tmp = memchr (ptr, '\0', precision)) ? tmp - ptr : precision;

	    /* o */
	    count += format_last (bufp, sizep, ptr, flag, '\0', width, size, size);
	    throw (s);
	    break;

	/*
	 * ȊO....
	 */

	default:
	    fputc (*fmt, fp);
	    count++;
	    break;
	}
    }
}

/* File scope functions */
static char *format_L (unsigned long long uval, int flag,
		       int precision, int base, int *size, char *charset)
{
    char *ptr;
    unsigned int shift, mask;
    static char buffer[64];

    /* ϊpobt@̍ŌɃ|C^ړ */
    ptr = buffer + sizeof buffer;

    /* `FbN */
    switch (base) {
    case 2:
	shift = 1;
	mask = 1;
	break;
    case 8:
	shift = 3;
	mask = 7;
	break;
    case 16:
	shift = 4;
	mask = 15;
	break;
    default:
	shift = 0;
	mask = 0;
	break;
    }

    /* 𕶎ɕϊ */
    if (precision != 0) {

	/* Px */
	{

	    unsigned long udivs;
	    unsigned long uvals;

	    /* Pxɕϊ */
	    uvals = (unsigned long) uval;

	    /* VtgŌvZ\ */
	    if (shift > 0) {
		do {
		    *--ptr = charset[uvals & mask];
		    uvals = uvals >> shift;
		} while (uvals > 0UL);
	    }

	    /* Vtgł͌vZłȂ */
	    else {
		do {
		    udivs = uvals / base;
		    *--ptr = charset[uvals - udivs * base];
		    uvals = udivs;
		} while (uvals > 0UL);
	    }

	}

    }

    /* ϊ̃TCYݒ */
    *size = buffer + sizeof buffer - ptr;

    /* ̐擪AhXԂ */
    return ptr;
}

/* File scope functions */
static int format_last (char** bufp, int* sizep, char *ptr, int flag,
			int sign, int width, int precision, int bodysize)
{
    int n, real_size;
    int string_size;
    unsigned char *p = ptr;

    /* ̒𒲐 */
    string_size = (bodysize
		   + ((sign) ? 1 : 0)
		   + ((flag & _P_OCTAPFIX) ? 1 : 0)
		   + ((flag & (_P_LHEXPFIX | _P_UHEXPFIX | _P_BINAPFIX)) ? 2 : 0));

    /* xĒ𒲐 */
    real_size = (precision > string_size) ? precision : string_size;

    /* 1. El߂̂߂̋󔒂o͂ */
    if ((flag & (_P_RUGLEFT | _P_ZEROPAD)) == 0)
	for (n = real_size; n < width; n++)
	    fputc (' ', fp);

    /* 2. o */
    if (sign)
	fputc (sign, fp);

    /* 2. 邢2i8i16ivtBNX */
    else if (flag & (_P_OCTAPFIX | _P_LHEXPFIX | _P_UHEXPFIX | _P_BINAPFIX)) {
	fputc ('0', fp);
	if (flag & _P_LHEXPFIX)
	    fputc ('x', fp);
	else if (flag & _P_UHEXPFIX)
	    fputc ('X', fp);
	else if (flag & _P_BINAPFIX)
	    fputc ('b', fp);
    }

    /* 3. El߂̂߂Ƀ[l߂sȂ */
    if ((flag & (_P_RUGLEFT | _P_ZEROPAD)) == _P_ZEROPAD)
	for (n = real_size; n < width; n++)
	    fputc ('0', fp);

    /* 4. xp̃[l߂sȂ */
    for (n = string_size; n < precision; n++)
	fputc ('0', fp);

    /* 5. lA邢͕{̂o͂ */
    for (n = bodysize; --n >= 0; )
	fputc (*p++, fp);

    /* 6. lߗp̋󔒒sȂ */
    if (flag & _P_RUGLEFT)
	for (n = real_size; n < width; n++)
	    fputc (' ', fp);

    /* o͕Ԃ */
    return (width > real_size) ? width : real_size;
}

/* File scope functions */
static void append (int c, char** bufp, int* sizep)
{
  *bufp = Realloc (*bufp, *sizep + 2);
  (*bufp)[(*sizep)++] = c;
}


/* EOF */
