/* pmmi.c - Portable MMI
 * Copyright (C) 2000 Tachibana Eriko
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


/* System headers */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mbctype.h>
#include <sys/dos.h>

/* Local headers */
#include "pmmi.h"


/* Macros */
#define eq(s1, s2)	(strcmp (s1, s2) == 0)


/* File scope function */
static void throw_and_null (Straw** s_ptr);
static void __volatile error_exit (const char* mes);
static void check_div0 (int n);


/* File scope variables */
static const char program[] = "pmmi";
static const char version[] = "0.90";
static const char date[] = "2000/03/11";
static const char author[] = "Ԃq";

static const char* argv0;


/* File scope function */
static void
print_usage (void)
{
  printf ("%s version %s  %s %s.\n", program, version, date, author);
  printf ("usage: %s <mmi> ...\n", argv0);
}

/* File scope function */
static void
print_version (void)
{
  printf ("%s version %s\n", program, version);
}


/* File scope function */
static int
exec_mmi (char** argptr)
{
  char* p;
  static Straw* strtok_ptr;

  while ((p = *argptr++) != NULL)
    {
       /* {v~eBu */
      if (eq (p, "drop"))
	{
	  throw (pop ());		/* top ̂Ă */
	}
      else if (eq (p, "dup"))
	{
	  Straw* s = pop ();
	  push (s->str);		/* top ɖ߂ */
	  push (s->str);		/* ̂ς */
	  throw (s);
	}
      else if (eq (p, "over"))
	{
	  Straw* s1 = pop ();
	  Straw* s2 = pop ();
	  push (s2->str);		/* 2nd ɖ߂ */
	  push (s1->str);		/* top ɖ߂ */
	  push (s2->str);		/* 2nd Ɠ̂ς */
	  throw (s2);
	  throw (s1);
	}
      else if (eq (p, "2over"))
	{
	  Straw* s1 = pop ();
	  Straw* s2 = pop ();
	  Straw* s3 = pop ();
	  push (s3->str);		/* 3rd ɖ߂ */
	  push (s2->str);		/* 2nd ɖ߂ */
	  push (s1->str);		/* top ɖ߂ */
	  push (s3->str);		/* 3rd Ɠ̂ς */
	  throw (s3);
	  throw (s2);
	  throw (s1);
	}
      else if (eq (p, "swap"))
	{
	  Straw* s1 = pop ();
	  Straw* s2 = pop ();
	  push (s1->str);		/* top ς */
	  push (s2->str);		/* 2nd ς */
	  throw (s2);
	  throw (s1);
	}
      else if (eq (p, "2swap"))
	{
	  Straw* s1 = pop ();
	  Straw* s2 = pop ();
	  Straw* s3 = pop ();
	  push (s1->str);		/* top ς */
	  push (s2->str);		/* 2nd ɖ߂ */
	  push (s3->str);		/* 3rd ς */
	  throw (s3);
	  throw (s2);
	  throw (s1);
	}

      /* o̓v~eBu */
      else if (eq (p, "pr"))
	{
	  Straw* s = pop ();
	  fputs (s->str, stdout);
	  throw (s);
	}
      else if (eq (p, "puts"))
	{
	  Straw* s = pop ();
	  puts (s->str);
	  throw (s);
	}
      else if (eq (p, "setenv"))
	{
	  Straw* name = pop ();
	  Straw* s = pop ();
	  _dos_setenv (name->str, NULL, s->str);
	  throw (name);
	  throw (s);
	}

      /* lZv~eBu */
      else if (eq (p, "+"))
	{
	  int top = popi ();
	  pushi (popi () + top);
	}
      else if (eq (p, "-"))
	{
	  int top = popi ();
	  pushi (popi () - top);
	}
      else if (eq (p, "*"))
	{
	  int top = popi ();
	  pushi (popi () * top);
	}
      else if (eq (p, "/"))
	{
	  int top = popi ();
	  check_div0 (top);
	  pushi (popi () / top);
	}
      else if (eq (p, "%"))
	{
	  int top = popi ();
	  check_div0 (top);
	  pushi (popi () % top);
	}

      /* _Zv~eBu */
      else if (eq (p, "and"))
	{
	  int top = popi ();
	  pushi (popi () & top);
	}
      else if (eq (p, "or"))
	{
	  int top = popi ();
	  pushi (popi () | top);
	}
      else if (eq (p, "xor"))
	{
	  int top = popi ();
	  pushi (popi () ^ top);
	}
      else if (eq (p, "not"))
	{
	  pushi (~popi ());
	}

      /* 擾v~eBu */
      else if (eq (p, "rnd"))
	{
	  pushi (random () >> (LONG_BIT - 16));
	}
      else if (eq (p, "rand"))
	{
	  pushi (((popi () + 1.0) / (double) ULONG_MAX) * random ());
	}
      else if (eq (p, "srand"))
	{
	  srand (popi ());
	}

      /* R[hϊv~eBu */
      else if (eq (p, "asc"))
	{
	  char buf[32];
	  Straw* s = pop ();
	  snprintf (buf, sizeof buf, "0x%02x", *(unsigned char*)(s->str));
	  throw (s);
	  push (buf);
	}
      else if (eq (p, "chr"))
	{
	  char buf[2];
	  buf[0] = popi () & 0xff;
	  buf[1] = '\0';
	  push (buf);
	}

      /* 񑀍v~eBu */
      else if (eq (p, "strmid"))
	{
	  int tail = popi ();
	  int from = popi ();
	  Straw* s = pop ();
	  int len = strlen (s->str);

	  tail += from;
	  from = min (from, len);
	  tail = min (tail, len);
	  s->str[tail] = '\0';
	  push (s->str + from);

	  throw (s);;
	}
      else if (eq (p, "strfmt"))
	{
	  extern void doprnt (const char *format, char** bufp, int* sizep);
	  Straw* top = pop ();
	  char* buf = NULL;
	  int size = 0;

	  doprnt (top->str, &buf, &size);
	  throw (top);
	  push (buf ? buf : "");
	  free (buf);
	}
      else if (eq (p, "strlen"))
	{
	  Straw* s = pop ();
	  pushi (strlen (s->str));
	  throw (s);;
	}
      else if (eq (p, "instr"))
	{
	  int from = popi ();
	  Straw* s2 = pop ();
	  Straw* s3 = pop ();
	  int len = strlen (s3->str);
	  char* p;

	  if (from < len && (p = strstr (s3->str + from, s2->str)) != NULL)
	    pushi (p - s3->str);
	  else
	    pushi (0);
	  throw (s3);
	  throw (s2);
	}
      else if (eq (p, "strtok"))
	{
	  Straw* delim = pop ();
	  Straw* s2 = pop ();
	  char* p = strtok (s2->str, delim->str);

	  throw_and_null (&strtok_ptr);
	  push (p ? p : "");
	  strtok_ptr = s2;
	  throw (delim);
	}
      else if (eq (p, "nstrtok"))
	{
	  Straw* delim = pop ();
	  char* p = strtok (NULL, delim->str);

	  if (p == NULL)
	    throw_and_null (&strtok_ptr);
	  push (p ? p : "");
	  throw (delim);
	}
      else if (eq (p, "lstrtok"))
	{
	  Straw* delim = pop ();
	  Straw* s2 = pop ();
	  char* p = strtok (s2->str, delim->str);
	  int n;

	  throw_and_null (&strtok_ptr);
	  for (n = 0; p; p = strtok (NULL, delim->str))
	    {
	      push (p);
	      ++n;
	    }
	  throw (delim);
	  throw (s2);
	  pushi (n);
	}

      /* 񑀍v~eBu(Z^On) */
      else if (eq (p, "StrCenter"))
	{
	  int top = popi ();
	  Straw* s2 = pop ();
	  int len = strlen (s2->str);
	  int spc = (top - len) / 2;

	  if (spc <= 0)
	    push (s2->str);		/* 2nd ̂܂ܐς */
	  else
	    {
	      char* p = Malloc (spc + len + 1);
	      memset (p, ' ', spc);
	      strcpy (p + spc, s2->str);
	      push (p);
	      free (p);
	    }
	  throw (s2);
	}
      else if (eq (p, "StrRight"))
	{
	  int top = popi ();
	  Straw* s2 = pop ();
	  int len = strlen (s2->str);
	  int spc = (top - len);

	  if (spc <= 0)
	    push (s2->str);		/* 2nd ̂܂ܐς */
	  else
	    {
	      char* p = Malloc (spc + len + 1);
	      memset (p, ' ', spc);
	      strcpy (p + spc, s2->str);
	      push (p);
	      free (p);
	    }
	  throw (s2);
	}
      else if (eq (p, "StrLeft"))
	{
	  int top = popi ();
	  Straw* s2 = pop ();
	  int len = strlen (s2->str);
	  int spc = (top - len);

	  if (spc <= 0)
	    push (s2->str);		/* 2nd ̂܂ܐς */
	  else
	    {
	      char* p = Malloc (len + spc + 1);
	      strcpy (p, s2->str);
	      memset (p + len, ' ', spc);
	      p[len + spc] = '\0';
	      push (p);
	      free (p);
	    }
	  throw (s2);
	}

      /* z񑀍v~eBu */
      else if (eq (p, "assign"))
	{
	  int top = popi ();
	  char** p = Malloc (sizeof (int) * (top + 1));
	  int i;

	  pushi ((int) p);
	  *p++ = (char*) top;		/* vf */
	  for (i = 0; i < top; ++i)
	    *p++ = NULL;
	}
      else if (eq (p, "dispose"))
	{
	  char** ptr = (char**) popi ();
	  char** p = ptr;
	  int n = (int) *p++;		/* vf */
	  int i;

	  for (i = 0; i < n; ++i)
	    {
	      char* s = *p++;
	      if (s)
		free (s);
	    }
	  free (ptr);
	}
      else if (eq (p, "setarray"))
	{
	  Straw* top = pop ();
	  int n = popi ();
	  char** p = (char**) popi ();
	  char* buf = Malloc (strlen (top->str) + 1);

	  strcpy (buf, top->str);
	  if (p[1 + n])
	    free (p[1 + n]);
	  p[1 + n] = buf;
	  throw (top);
	}
      else if (eq (p, "getarray"))
	{
	  int n = popi ();
	  char** p = (char**) popi ();

	  if ((int) p[0] <= n)
	    error_exit ("z̗vfԍ͈͊Oł.");
	  push (p[1 + n] ? p[1 + n] : "");
	}

      /* t@Cv~eBu */
      else if (eq (p, "fopen"))
	{
	  Straw* mode = pop ();
	  Straw* name = pop ();
	  FILE* fp = fopen (name->str, mode->str);
	  throw (name);
	  throw (mode);
	  pushi (fp ? (int) fp : -1);
	}
      else if (eq (p, "fclose"))
	{
	  FILE* fp = (FILE*) popi ();
	  fclose (fp);
	}
      else if (eq (p, "fgets"))
	{
	  FILE* fp = (FILE*) popi ();
	  char* buf = NULL;
	  int len = 0;
	  int c;

	  while ((c = fgetc (fp)) != EOF)
	    {
	      buf = Realloc (buf, len + 2);
	      buf[len++] = c;
	      if (c == '\n')
		break;
	    }
	  if (len)
	    {
	      buf[len] = '\0';
	      push (buf);
	      free (buf);
	    }
	  else
	    push ("\x1a");
	}
      else if (eq (p, "fputs"))
	{
	  FILE* fp = (FILE*) popi ();
	  Straw* s = pop ();
	  fputs (s->str, fp);
	  throw (s);
	}
      else if (eq (p, "fgetc"))
	{
	  FILE* fp = (FILE*) popi ();
	  int c = fgetc (fp);
	  pushi (c == EOF ? '\x1a' : c);
	}
      else if (eq (p, "fputc"))
	{
	  FILE* fp = (FILE*) popi ();
	  fputc (popi (), fp);
	}
      else if (eq (p, "fseek"))
	{
	  FILE* fp = (FILE*) popi ();
	  int mode = popi ();
	  long offs = popi ();
	  fseek (fp, offs, mode);
	}

      /* /lv~eBu */
      else
	{
	  char* s = p;
	  char* d = p;
	  unsigned char c;
	  int quote = 0;

	  while ((c = *s++) != '\0')
	    {
	      if (ismbblead (c) && *s)
		{
		  *d++ = c;
		  *d++ = *s++;
		}
	      else if (c == '"')
		{
		  quote ^= 1;
		}
	      else if (c == '\\')
		{
		  unsigned char c2 = *s++;
		  unsigned char h1, h2;

		  if (c2 == 'n' || c2 == 'l')
		    *d++ = '\x0a';
		  else if (c2 == 'r')
		    *d++ = '\x0d';
		  else if (c2 == 'c')
		    *d++ = '\x03';
		  else if (c2 == 'g')
		    *d++ = '\x07';
		  else if (c2 == 't')
		    *d++ = '\x09';
		  else if (c2 == 'e')
		    *d++ = '\x1b';
		  else if (c2 == 'h')
		    *d++ = '\x08';
		  else if (c2 == 'z')
		    *d++ = '\x1a';
		  else if (c2 == 'x'
			&& isxdigit (h1 = tolower ((unsigned char) s[0]))
			&& isxdigit (h2 = tolower ((unsigned char) s[1])))
		    {
		      int n = (((h1 <= '9') ? (h1 - '0') : (h1 - 'a' + 10)) << 4)
			    +  ((h2 <= '9') ? (h2 - '0') : (h2 - 'a' + 10));
		      s += 2;
		      *d++ = n;
		    }
		  else
		    *d++ = c;
		}
	      else
		*d++ = c;
	    }
	  *d = '\0';

	  push (p);
	}
    }

  return EXIT_SUCCESS;
}

/* File scope function */
static void
throw_and_null (Straw** s_ptr)
{
  Straw* s = *s_ptr;

  if (s)
    {
      throw (s);
      *s_ptr = NULL;
    }
}


/* Function */
int
main (int argc, char* argv[])
{

  argv0 = *argv++;
  if (argc == 1)
    {
      print_usage ();
      return EXIT_FAILURE;
    }
  else if (argc == 2)
    {
      char* p = *argv;

      if (eq (p, "-h") || eq (p, "-H") || eq (p, "-?") || eq (p, "--help"))
	{
	  print_usage ();
	  return EXIT_SUCCESS;
	}
      else if (eq (p, "-v") || eq (p, "-V") || eq (p, "--version"))
	{
	  print_version ();
	  return EXIT_SUCCESS;
	}
    }

  return exec_mmi ((char**) argv);
}



/* -------- X^bN -------- */

/* File scope variables */
static Straw* stack_ptr;
static int stack_num;


/* Function */
void
pushi (int n)
{
  char buf[32];

  snprintf (buf, sizeof buf, "%d", n);
  push (buf);
}

/* Function */
void
push (char* str)
{
  Straw* s = Malloc (sizeof (Straw));

  s->prev = stack_ptr;
  s->str = Malloc (strlen (str) + 1);
  strcpy (s->str, str);

  stack_ptr = s;
  ++stack_num;
}

/* Function */
int
popi (void)
{
  Straw* s = pop ();
  char* endptr;
  int n;

  n = (int) strtol (s->str, &endptr, 0);
  if (endptr == s->str || *endptr)
    error_exit ("l̕\܂.");
  throw (s);
  return n;
}

/* Function */
Straw*
pop (void)
{
  Straw* s;

  if (--stack_num < 0)
    error_exit ("X^bN͋ł.");

  s = stack_ptr;
  stack_ptr = s->prev;

  return s;
}

/* Function */
void
throw3 (Straw* s1, Straw* s2, Straw* s3)
{
  throw (s1);
  throw (s2);
  throw (s3);
}

/* Function */
void
throw2 (Straw* s1, Straw* s2)
{
  throw (s1);
  throw (s2);
}

/* Function */
void
throw (Straw* s)
{
  free (s->str);
  free (s);
}


/* -------- Ǘ -------- */

void*
Malloc (int size)
{
  void* p = malloc (size);

  if (p == NULL)
    error_exit ("܂.");
  return p;
}

void*
Realloc (void* ptr, int size)
{
  void* p = realloc (ptr, size);

  if (p == NULL)
    error_exit ("܂.");
  return p;
}


/* -------- G[ -------- */

/* File scope function */
static void
check_div0 (int n)
{
  if (n == 0)
    error_exit ("0 ŏZ܂.");
}

/* File scope function */
static void __volatile
error_exit (const char* mes)
{
  fprintf (stderr, "%s: %s\n", program, mes);
  exit (EXIT_FAILURE);
}


/* EOF */
